#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include "display.h"                                                                                                   
#define ESCAPE 27
#ifdef HAVE_OPENGL
#define NCOLORS 1000
#else
#ifdef HAVE_MPE
#define NCOLORS 256
#else
#define NCOLORS 1000
#endif
#endif

#define WINSIZE 600
#define TEXTHEIGHT 18

int DrawString(float x, float y, float z, char* string);
void InitGL(int width, int height);
void DrawSquares(void);
void DrawBoxes(void);
void SelectionRec(void);
void mouseClick(int button, int state, int x, int y);
void mouseDrag(int x, int y);
void keyPressed(unsigned char key, int x, int y);
void Scale();
void mpe_main_loop(void);
void display_get_event(void);

struct ColorTable {
   float Red;
   float Green;
   float Blue;
};

static int autoscale = 0;
static double scaleMax = 100.0, scaleMin = 0.0;
static double graScale = 11.0;
static char display_label[200]="\0";

static int cell_centered = 0;

static double display_circle_radius=-1.0;

#ifndef HAVE_MPI
#ifdef HAVE_MPE
static int MPI_COMM_WORLD=0;
#endif
#endif

#ifdef HAVE_OPENGL
static struct ColorTable Rainbow[NCOLORS];
static int window;
static int contour_window;
static int graph_window;
static int scale_window;
static int xcut_window;
static int status_window;
#endif
#ifdef HAVE_MPE
static MPE_Color *color_array;
static int ncolors=256;
static MPE_XGraph window;
static double xconv = 0.0;
static double yconv = 0.0;
static XFontStruct *font_info;
void (*idlefunction)(void);
#endif

void (*keyboard_function)(unsigned char key, int x, int y);
void (*mouse_click_function)(int button, int state, int x, int y);

#ifdef HAVE_OPENGL
static  double xstart, ystart, xend, yend;
#endif
enum mode_choice {EYE, MOVE, DRAW};
static int mode = MOVE;

static int width, height;
static int contourHeight;
static int contourWidth;
static int graphHeight;
static int graphWidth;
static int scaleHeight;
static int scaleWidth;
static int xcutHeight;
static int xcutWidth;
static int statusHeight;
static int statusWidth;

static  double display_xmin=0.0, display_xmax=0.0, display_ymin=0.0, display_ymax=0.0;

#ifdef HAVE_OPENGL
static GLfloat xrot = 0.0, yrot = 0.0, xloc = 0.0, zloc = 0.0;
#else
#ifdef HAVE_MPE
static double xrot = 0.0, yrot = 0.0, xloc = 0.0, zloc = 0.0;
#else
static double xrot = 0.0, yrot = 0.0, xloc = 0.0, zloc = 0.0;
#endif
#endif

static int display_outline;
static int display_view_mode = 0;
static int display_isize    = 0;
static int display_jsize    = 0;
static int display_myoffset = 0;
static double *x=NULL, *y=NULL, *dx=NULL, *dy=NULL;
static double **data=NULL;
static int **display_proc=NULL;
static int rank = 0;

int DrawString(float x, float y, float z, char* string) {
#ifdef HAVE_OPENGL
   char *c;
   glColor3f(0.0f, 0.0f, 0.0f);
   glRasterPos3f(x, y, z);
   //int count = 0;
   for(c = string; *c != '\0'; c++) {
      //glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_10, *c);
      //glutBitmapCharacter(GLUT_BITMAP_HELVETICA_18, *c);
      //glutBitmapCharacter(GLUT_BITMAP_HELVETICA_12, *c);
      //glutBitmapCharacter(GLUT_BITMAP_9_BY_15, *c);
      glutBitmapCharacter(GLUT_BITMAP_8_BY_13, *c);
      //printf("printing %d char %c\n",count,*c);
      //count++;
   }
   //printf("Printed %d characters\n",count);
#endif
#ifdef HAVE_MPE
#endif
   return 1;
}
#ifdef HAVE_OPENGL
void InitGL(int width, int height) {
   glClearColor(1.0f, 1.0f, 1.0f, 1.0f);
   glDepthFunc(GL_LESS);
   glShadeModel(GL_SMOOTH);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();
}
#endif
void init_display(int *argc, char **argv, const char *title, int rank_in){
   if (rank_in) rank = rank_in;

   width = (WINSIZE / (display_ymax - display_ymin)) * (display_xmax - display_xmin);
   height = WINSIZE;

   contourHeight = height-height/5;
   contourWidth  = width-50;
   graphHeight   = height-contourHeight-TEXTHEIGHT;
   graphWidth    = width-50;
   scaleHeight   = height-contourHeight;
   scaleWidth    = 50;
   xcutHeight    = contourHeight;
   xcutWidth     = 50;

   statusHeight  = TEXTHEIGHT;
   statusWidth   = width-50;

#ifdef HAVE_OPENGL
   glutInit(argc, argv);
   glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
   if (rank == 0) {
      glutInitWindowSize(width, WINSIZE);
      glutInitWindowPosition(20, 20);
   } else {
      glutInitWindowSize(1,1);
      glutInitWindowPosition(5, 5);
   }

   window = glutCreateWindow(title);
   glutDisplayFunc(&draw_scene);
   contour_window = glutCreateSubWindow(window, 0,            0,             contourWidth, contourHeight);
   glutDisplayFunc(&draw_contour);
   graph_window   = glutCreateSubWindow(window, 0,            contourHeight, graphWidth,   graphHeight);
   glutDisplayFunc(&draw_graph);
   scale_window   = glutCreateSubWindow(window, contourWidth, contourHeight, scaleWidth,   scaleHeight);
   glutDisplayFunc(&draw_scale);
   xcut_window    = glutCreateSubWindow(window, contourWidth, 0,             xcutWidth,    xcutHeight);
   glutDisplayFunc(&draw_xcut);
   status_window  = glutCreateSubWindow(window, 0,            contourHeight+graphHeight, statusWidth,  statusHeight);
   glutDisplayFunc(&draw_status);

   glutSetWindow(contour_window);
#endif

#ifdef HAVE_MPE
   char fontname[30];
   char *displayname=NULL;

   /* Open the graphics display */

   if (*argc > 2 && strcmp( argv[1], "-display" ) == 0) {
      displayname = (char *)malloc( strlen( argv[2] ) + 1 );
      strcpy(displayname, argv[2]);
   }

   if (displayname == NULL){
      displayname = getenv("DISPLAY");
      if (displayname == NULL){
         displayname = (char *)malloc((size_t)28);
         displayname = ":0";
      }
   }

#ifdef HAVE_MPI
   int mpi_init_flag=0;
   MPI_Initialized(&mpi_init_flag);
   if (mpi_init_flag){
      MPI_Comm_rank(MPI_COMM_WORLD, &rank);
   } else {
      MPI_Init(argc, &argv);
   }
#endif
   if (MPE_Open_graphics( &window, MPI_COMM_WORLD, displayname,
                         -1, -1, WINSIZE, WINSIZE, 0) != MPE_SUCCESS){
      printf("Problem opening graphics window\n");
      exit(-1);
   }

   xconv = (double)contourWidth/ (display_xmax-display_xmin);
   yconv = (double)contourHeight/(display_ymax-display_ymin);

   if (rank == 0){
      XSelectInput( window->xwin->disp, window->xwin->win, MPE_XEVT_IDLE_MASK |
                  ButtonPressMask   |
                  ButtonReleaseMask |
                  Button1MotionMask |
                  KeyPressMask      |
                  ExposureMask      |
                  StructureNotifyMask);
   }

   MPE_Update(window);

   // if we want to change fonts, we can load a new one with the following
   strcpy(fontname,"fixed");
   font_info = XLoadQueryFont( window->xwin->disp, fontname );
#endif

   Scale();

#ifdef HAVE_OPENGL
   glutDisplayFunc(&draw_scene);

   glutKeyboardFunc(&keyPressed);
   glutMouseFunc(&mouseClick);
   glutMotionFunc(&mouseDrag);
   InitGL(width, WINSIZE);
#endif
}

void draw_graph(void){
   int graRow = display_jsize/2;

#ifdef HAVE_OPENGL
   if (rank) return; // Only processor 0 draws to screen for OpenGL		 
   glutSetWindow(graph_window);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   //gluOrtho2D(display_xmin, scaleMin, display_xmax, scaleMax);
   double maxScale = (double)scaleMax*graphHeight/(graphHeight-TEXTHEIGHT);
   glOrtho(display_xmin, display_xmax, scaleMin, maxScale, 0, 1);

   glDisable(GL_DEPTH_TEST);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glTranslatef(0.375, 0.375, 0);

   glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

   glClear(GL_COLOR_BUFFER_BIT);

   glColor3f(0.0f, 0.0f, 0.0f);

   double step = NCOLORS/(maxScale - scaleMin);

   glBegin(GL_LINE_STRIP);
      for (int i=0; i<display_isize; i++){
         int color = (int)(data[graRow][i]-scaleMin)*step;
         if (color < 0) {
            color=0;
         }
         if (color >= NCOLORS) color = NCOLORS-1;
   
         glColor3f(Rainbow[color].Red, Rainbow[color].Green, Rainbow[color].Blue);
         if (cell_centered) {
            //printf("x %lf data %lf\n",x[i],data[graRow][i]);
            glVertex2f(x[i],data[graRow][i]);
         } else {
            //printf("x %lf data %lf\n",x[i]+0.5*dx[i],data[graRow][i]);
            glVertex2f(x[i]+0.5*dx[i],data[graRow][i]);
         }
      }
   glEnd();

   //glLoadIdentity();

   glutSwapBuffers();
#endif
#ifdef HAVE_MPE
   int xloc, xwid;

   //draw graph
   if (display_myoffset<=graRow && display_myoffset+display_jsize>graRow){
      MPE_Fill_rectangle(window, 0, contourHeight, graphWidth, graphHeight, MPE_WHITE);
      MPE_Fill_rectangle(window, 0, contourHeight, graphWidth, 2, MPE_BLACK);
      //MPE_Fill_rectangle(window, 0, (double)height-(double)2*TEXTHEIGHT-(double)graphHeight*((scaleMax-scaleMin)/graScale), graphWidth, 1, color_array[ncolors/2]);
      for(unsigned int i=0; i<display_isize; i++){
         if (cell_centered) {
            xloc = (int)((x[i]-0.5*dx[i]-display_xmin)*xconv        + 0.5);
            xwid = (int)((x[i]+0.5*dx[i]-display_xmin)*xconv - xloc + 0.5);
         } else {
            xloc = (int)((x[i]      -display_xmin)*xconv        + 0.5);
            xwid = (int)((x[i]+dx[i]-display_xmin)*xconv - xloc + 0.5);
         }
         if (xwid <= 0) continue;
         if (xloc >= contourWidth || xloc + xwid < 0) continue;

         if (xloc + xwid > contourWidth) xwid = contourWidth - xloc;
         if (xloc < 0) xloc = 0;
         int yloc = (double)height-(double)2*TEXTHEIGHT-(double)graphHeight*(data[graRow-display_myoffset][i]/graScale);
         if (yloc > contourHeight) MPE_Fill_rectangle(window, xloc, yloc, xwid, 1, color_array[1]);
      }
   }
#endif
}

void draw_scale(void){
#ifdef HAVE_OPENGL
   if (rank != 0) return;
   glutSetWindow(scale_window);

   glMatrixMode(GL_PROJECTION);
   glLoadIdentity();

   //gluOrtho2D(display_xmin, 0, display_xmax, NCOLORS);
   //glOrtho(0, 1, scaleMin, scaleMax, 0, 1);
   double range = (double)NCOLORS*scaleHeight/(scaleHeight-2*TEXTHEIGHT);
   float minScale = - 0.5*(range-(double)NCOLORS);
   float maxScale = (double)NCOLORS + 0.5*(range-(double)NCOLORS);
   glOrtho(0.0f, 1.0f, minScale, maxScale, 0.0f, 1.0f);

   glDisable(GL_DEPTH_TEST);
   glMatrixMode(GL_MODELVIEW);
   glLoadIdentity();

   glTranslatef(0.375, 0.375, 0);

   glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

   glClear(GL_COLOR_BUFFER_BIT);
   //int iscaleHeight=graphHeight*((scaleMax-scaleMin)/graScale);

   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glBegin(GL_QUADS);

   //double step = NCOLORS/(scaleMax - scaleMin);
   double step = 1.0;
   int ilast = 0;

   //draw color for scale
   for (unsigned int i=0; i<NCOLORS; i++) {
      int color = (int)((double)i*step);
      if (color < 0) {
         color=0;
      }
      if (color >= NCOLORS) color = NCOLORS-1;

      glColor3f(Rainbow[color].Red, Rainbow[color].Green, Rainbow[color].Blue);
      glVertex2f(-0.3f, (float)ilast);
      glVertex2f(-0.3f, (float)i);
      glVertex2f(1.0f, (float)i);
      glVertex2f(1.0f, (float)ilast);
      ilast=i;
   }

   glEnd(); 

   char maxGra[20];
   char maxCol[20];
   char min[20];
   sprintf( maxGra, "%6.4e", graScale);
   sprintf( maxCol, "%6.4e", scaleMax-scaleMin);
   sprintf( min, "%6.4e", 0.0);

   DrawString(-0.3,(float)NCOLORS+(float)NCOLORS/100.0f,0.0,maxGra);
   DrawString(-0.3,minScale+(float)NCOLORS/100.0f,0.0,min);

   glLoadIdentity();

   glutSwapBuffers();
#endif
#ifdef HAVE_MPE
   int plot_value;
   if(display_myoffset==0){
      //graph scale
      MPE_Fill_rectangle(window, graphWidth, contourHeight, scaleWidth, graphHeight-(graphHeight*(int)((scaleMax-scaleMin)/graScale) ), MPE_BLACK);
      int iscaleHeight=(graphHeight-2*TEXTHEIGHT)*((scaleMax-scaleMin)/graScale);
      for (unsigned int i=0; i<=iscaleHeight; i++) {
      //draw color for scale
        plot_value = ncolors - ((double)ncolors*(double)i/(double)iscaleHeight) - 1;
         //printf("plot_value %d %d %lf %lf %d\n", ncolors, i, iscaleHeight, (double)ncolors*(double)i/(double)iscaleHeight,  plot_value);
         MPE_Fill_rectangle(window, graphWidth, (int)(height-2*TEXTHEIGHT-i), scaleWidth, 1, color_array[plot_value]);
      }

      //graph legend
      char maxGra[20];
      char maxCol[20];
      char min[20];
      sprintf( maxGra, "%6.4e", graScale);
      sprintf( maxCol, "%6.4e", scaleMax-scaleMin);
      sprintf( min, "%6.4e", 0.0);
      MPE_Fill_rectangle(window, 0, height-2*TEXTHEIGHT, graphWidth+scaleWidth, 2*TEXTHEIGHT, MPE_WHITE);
      MPE_Fill_rectangle(window, graphWidth, 0, 2, height, MPE_BLACK);
      MPE_Fill_rectangle(window, 0, height-2*TEXTHEIGHT, width, 2, MPE_BLACK);

      MPE_Draw_string(window, graphWidth+5,contourHeight-3,MPE_BLACK,maxGra);
      if(graphHeight-(graphHeight*((scaleMax-scaleMin)/graScale))>TEXTHEIGHT) MPE_Draw_string(window, graphWidth+5, height-2*TEXTHEIGHT-(graphHeight*(int)((scaleMax-scaleMin)/graScale))-3, MPE_WHITE, maxCol);
      MPE_Draw_string(window, graphWidth+5,height-TEXTHEIGHT-3,MPE_BLACK,min);

   }
#endif
}

void draw_xcut(void){
#ifdef HAVE_OPENGL
   if (rank) return; // Only processor 0 draws to screen for OpenGL		 
   glutSetWindow(xcut_window);

   glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();

   //set up 2D viewing
   gluOrtho2D(display_xmin, display_xmax, display_ymin, display_ymax);
  
   int graRow = display_jsize/2;

   /*Draw filled cells with color set by state value*/
   glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
   glBegin(GL_QUADS);

   int color = NCOLORS/2;

   glColor3f(Rainbow[color].Red, Rainbow[color].Green, Rainbow[color].Blue);

   if (cell_centered) {
      glVertex2f(display_xmin, y[graRow]-0.5*dy[graRow]);
      glVertex2f(display_xmax, y[graRow]-0.5*dy[graRow]);
      glVertex2f(display_xmax, y[graRow]+0.5*dy[graRow]);
      glVertex2f(display_xmin, y[graRow]+0.5*dy[graRow]);
   } else {
      glVertex2f(display_xmin, y[graRow]           );
      glVertex2f(display_xmax, y[graRow]           );
      glVertex2f(display_xmax, y[graRow]+dy[graRow]);
      glVertex2f(display_xmin, y[graRow]+dy[graRow]);
   }

   glEnd();

   glLoadIdentity();

   glutSwapBuffers();
#endif
#ifdef HAVE_MPE
   int graRow = display_jsize/2;
   if (display_myoffset<=graRow && display_myoffset+display_jsize>graRow){
      int yloc = (int)((y[graRow]-0.5*dy[graRow]-display_ymin)*yconv        + 0.5);
      int ywid = (int)((y[graRow]+0.5*dy[graRow]-display_ymin)*yconv - yloc + 0.5);
      yloc = contourHeight - yloc -1;

      MPE_Fill_rectangle(window,contourWidth+2,yloc,xcutWidth,(ywid>1)?ywid:1,color_array[NCOLORS/2]);
   }
#endif
}

void draw_status(void){
#ifdef HAVE_OPENGL
   glutSetWindow(status_window);

   glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();
#endif

   //sprintf(display_label,"1234567890test\n");
   //printf("DEBUG string is %s\n",display_label);
   DrawString(-0.9,0.0,0.0,display_label);
}

void set_display_label(char *display_label_in){
   sprintf(display_label,"%s",display_label_in);
}

void set_keyboard_function(void (*function)(unsigned char, int, int)){
   keyboard_function = function;
}

void set_mouse_click_function(void (*function)(int, int, int, int)){
   mouse_click_function = function;
}

void set_idle_function(void (*function)(void)){
#ifdef HAVE_OPENGL
   glutIdleFunc(function);
#endif
#ifdef HAVE_MPE
   idlefunction = function;
#endif
}

void start_main_loop(void){
#ifdef HAVE_OPENGL
   glutMainLoop();
#endif
#ifdef HAVE_MPE
   mpe_main_loop();
#endif
}
   
void set_window(double display_xmin_in, double display_xmax_in, double display_ymin_in, double display_ymax_in){
   display_xmin = display_xmin_in;
   display_xmax = display_xmax_in;
   display_ymin = display_ymin_in;
   display_ymax = display_ymax_in;
}
void set_cell_data(double **data_in){
   data = data_in;
}
void set_cell_proc(int **display_proc_in){
   display_proc = display_proc_in;
}
void set_cell_coordinates(double *x_in, double *dx_in, double *y_in, double *dy_in){
   x = x_in;
   dx = dx_in;
   y = y_in;
   dy = dy_in;
}
void free_display(void){
#ifdef HAVE_OPENGL
   glutDestroyWindow(window);
#endif
}
void DisplayState(void) {
   int i, j, color;
   //vector<> &H = state->H;

   if (autoscale) {
      scaleMax=-1.0e30;
      scaleMin=1.0e30;
      for(j = 0; j<display_isize; j++) {
         for(i = 0; i<display_isize; i++) {
            if (data[j][i] > scaleMax) scaleMax = data[j][i];
            if (data[j][i] < scaleMin) scaleMin = data[j][i];
         }
      }
   }

   double step = NCOLORS/(scaleMax - scaleMin);
  
#ifdef HAVE_OPENGL
   //set up 2D viewing
   gluOrtho2D(display_xmin, display_xmax, display_ymin, display_ymax);
  
   for(j = 0; j < display_jsize; j++) {
      for(i = 0; i < display_isize; i++) {
         /*Draw filled cells with color set by state value*/
         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
         glBegin(GL_QUADS);

         color = (int)(data[j][i]-scaleMin)*step;
         if (color < 0) {
            color=0;
         }
         if (color >= NCOLORS) color = NCOLORS-1;
   
         glColor3f(Rainbow[color].Red, Rainbow[color].Green, Rainbow[color].Blue);
   
         if (cell_centered) {
            glVertex2f(x[i]-0.5*dx[i], y[j]-0.5*dy[j]);
            glVertex2f(x[i]+0.5*dx[i], y[j]-0.5*dy[j]);
            glVertex2f(x[i]+0.5*dx[i], y[j]+0.5*dy[j]);
            glVertex2f(x[i]-0.5*dx[i], y[j]+0.5*dy[j]);
         } else {
            glVertex2f(x[i],       y[j]);
            glVertex2f(x[i]+dx[i], y[j]);
            glVertex2f(x[i]+dx[i], y[j]+dy[j]);
            glVertex2f(x[i],       y[j]+dy[j]);
            //printf("vertex %d %d %lf %lf %lf %lf\n",j,i,x[i],y[i],x[i]+dx[i],y[i]+dx[i]);
         }
         glEnd();
   
         /*Draw cells again as outline*/
         if (display_outline) {
            glColor3f(0.0f, 0.0f, 0.0f);
            glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
            glBegin(GL_QUADS);
            if (cell_centered) {
               glVertex2f(x[i]-0.5*dx[i], y[j]-0.5*dy[j]);
               glVertex2f(x[i]+0.5*dx[i], y[j]-0.5*dy[j]);
               glVertex2f(x[i]+0.5*dx[i], y[j]+0.5*dy[j]);
               glVertex2f(x[i]-0.5*dx[i], y[j]+0.5*dy[j]);
            } else {
               glVertex2f(x[i],       y[j]);
               glVertex2f(x[i]+dx[i], y[j]);
               glVertex2f(x[i]+dx[i], y[j]+dy[j]);
               glVertex2f(x[i],       y[j]+dy[j]);
            }
            glEnd();
         }
      }
   }
#endif
#ifdef HAVE_MPE
   int xloc, xwid, yloc, ywid;
   int xloc1, xloc2, yloc1, yloc2;
   for(j = 0; j < display_jsize; j++) {
      int jj = j+display_myoffset;
      for(i = 0; i < display_isize; i++) {
         color = (int)(data[j][i]-scaleMin)*step;
         color = NCOLORS-color;
         if (color < 0) {
            color=0;
         }
         if (color >= NCOLORS) color = NCOLORS-1;

         //printf("DEBUG mesh -- xconv is %lf xmin %lf x[%d]=%lf\n",xconv,display_xmin,i,x[i]);
         //printf("DEBUG mesh -- yconv is %lf ymin %lf ymax %lf y[%d]=%lf\n",yconv,display_ymin,display_ymax,j,y[j]);
         if (cell_centered) {
            xloc = (int)((x[i]-0.5*dx[i]-display_xmin)*xconv);
            xwid = (int)((x[i]+0.5*dx[i]-display_xmin)*xconv-xloc);
            yloc = (int)((display_ymax-(y[jj]+0.5*dy[jj]))*yconv);
            ywid = (int)((display_ymax-(y[jj]-0.5*dy[jj]))*yconv);
         } else {
            xloc = (int)((x[i]      -display_xmin)*xconv);
            xwid = (int)((x[i]+dx[i]-display_xmin)*xconv-xloc);
            yloc = (int)((display_ymax-(y[jj]+dy[jj]))*yconv);
            ywid = (int)((display_ymax- y[jj]        )*yconv);
         }
         ywid -= yloc;
         //printf("xloc is %d xwid %d yloc is %d ywid %d color is  %d step is %d\n",xloc,xwid,yloc,ywid,color,step);
         MPE_Fill_rectangle(window, xloc, yloc, xwid, ywid, color_array[color]);

         if (display_outline) {
            if (cell_centered) {
               xloc1 = (int)((x[i]-0.5*dx[i]-display_xmin)*xconv);
               xloc2 = (int)((x[i]+0.5*dx[i]-display_xmin)*xconv);
               yloc1 = (int)((display_ymax-(y[jj]-0.5*dy[jj]))*yconv);
               yloc2 = (int)((display_ymax-(y[jj]+0.5*dy[jj]))*yconv);
            } else {
               xloc1 = (int)((x[i]      -display_xmin)*xconv);
               xloc2 = (int)((x[i]+dx[i]-display_xmin)*xconv);
               yloc1 = (int)((display_ymax- y[jj]        )*yconv);
               yloc2 = (int)((display_ymax-(y[jj]+dy[jj]))*yconv);
            }
            //printf("xloc1 is %d xloc2 %d yloc1 is %d yloc2 %d\n",xloc1,xloc2,yloc1,yloc2);
            MPE_Draw_line(window, xloc1, yloc2, xloc2, yloc2, MPE_BLACK);
            MPE_Draw_line(window, xloc1, yloc1, xloc2, yloc1, MPE_BLACK);
            MPE_Draw_line(window, xloc1, yloc1, xloc1, yloc2, MPE_BLACK);
            MPE_Draw_line(window, xloc2, yloc1, xloc2, yloc2, MPE_BLACK);
         }
      }
   }
   MPE_Update(window);
#endif

  
#ifdef HAVE_OPENGL
   glColor3f(0.0f, 0.0f, 0.0f);

   /* Draw circle for initial conditions */
   if (display_circle_radius > 0.0) {
      double radians;
      glBegin(GL_LINE_LOOP);
      for (double angle = 0.0; angle <= 360.0; angle+=1.0) {
         radians=angle*3.14159/180.0;
         glVertex2f(sin(radians) * display_circle_radius,cos(radians) * display_circle_radius);
      }
      glEnd();
   }
#endif
}

void DrawSquares(void) {
   int i, j, color, step;
   if (display_proc == NULL) return;
   step = NCOLORS/(display_proc[display_jsize-1][display_isize-1]+1);
   //step utilizes whole range of color, assumes last element of proc is last processor

#ifdef HAVE_OPENGL
   gluOrtho2D(display_xmin, display_xmax, display_ymin, display_ymax);
   //set up 2D viewing

   for(j = 0; j < display_jsize; j++) {
      for(i = 0; i < display_isize; i++) {
         /*Draw filled cells with color set by processor it is assigned to*/
         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
         glBegin(GL_QUADS);
            color = display_proc[j][i]*step;
            color = 10;
            glColor3f(Rainbow[color].Red, Rainbow[color].Green, Rainbow[color].Blue);

            if (cell_centered) {
               glVertex2f(x[i]-0.5*dx[i], y[j]-0.5*dy[j]);
               glVertex2f(x[i]+0.5*dx[i], y[j]-0.5*dy[j]);
               glVertex2f(x[i]+0.5*dx[i], y[j]+0.5*dy[j]);
               glVertex2f(x[i]-0.5*dx[i], y[j]+0.5*dy[j]);
            } else {
               glVertex2f(x[i],       y[j]);
               glVertex2f(x[i]+dx[i], y[j]);
               glVertex2f(x[i]+dx[i], y[j]+dy[j]);
               glVertex2f(x[i],       y[j]+dy[j]);
            }
         glEnd();

         /*Draw cells again as outline*/
         glColor3f(0.0f, 0.0f, 0.0f);
         glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
         glBegin(GL_QUADS);
            if (cell_centered) {
               glVertex2f(x[i]-0.5*dx[i], y[j]-0.5*dy[j]);
               glVertex2f(x[i]+0.5*dx[i], y[j]-0.5*dy[j]);
               glVertex2f(x[i]+0.5*dx[i], y[j]+0.5*dy[j]);
               glVertex2f(x[i]-0.5*dx[i], y[j]+0.5*dy[j]);
            } else {
               glVertex2f(x[i],       y[j]);
               glVertex2f(x[i]+dx[i], y[j]);
               glVertex2f(x[i]+dx[i], y[j]+dy[j]);
               glVertex2f(x[i],       y[j]+dy[j]);
            }
         glEnd();
      }
   }
#endif
#ifdef HAVE_MPE
   int xloc, xwid, yloc, ywid;
   int xloc1, xloc2, yloc1, yloc2;
   for(j = 0; j < display_jsize; j++) {
      for(i = 0; i < display_isize; i++) {
         //printf("DEBUG mesh -- xconv is %lf xmin %lf x[%d]=%lf\n",xconv,display_xmin,i,x[i]);
         if (cell_centered) {
            xloc = (int)((x[i]-0.5*dx[i]-display_xmin)*xconv);
            xwid = (int)((x[i]+0.5*dx[i]-display_xmin)*xconv-xloc);
            yloc = (int)((display_ymax-(y[j]+0.5*dy[j]))*yconv);
            ywid = (int)((display_ymax-(y[j]-0.5*dy[j]))*yconv);
         } else {
            xloc = (int)((x[i]      -display_xmin)*xconv);
            xwid = (int)((x[i]+dx[i]-display_xmin)*xconv-xloc);
            yloc = (int)((display_ymax-(y[j]+dy[j]))*yconv);
            ywid = (int)((display_ymax- y[j]       )*yconv);
         }
         ywid -= yloc;
         color = display_proc[j][i]*step;
         //printf("xloc is %d xwid %d yloc is %d ywid %d color is  %d step is %d\n",xloc,xwid,yloc,ywid,color,step);
         MPE_Fill_rectangle(window, xloc, yloc, xwid, ywid, color_array[color]);

         if (cell_centered) {
            xloc1 = (int)((x[i]-0.5*dx[i]-display_xmin)*xconv);
            xloc2 = (int)((x[i]+0.5*dx[i]-display_xmin)*xconv);
            yloc1 = (int)((display_ymax-(y[j]-0.5*dy[j]))*yconv);
            yloc2 = (int)((display_ymax-(y[j]+0.5*dy[j]))*yconv);
         } else {
            xloc1 = (int)((x[i]      -display_xmin)*xconv);
            xloc2 = (int)((x[i]+dx[i]-display_xmin)*xconv);
            yloc1 = (int)((display_ymax- y[j]       )*yconv);
            yloc2 = (int)((display_ymax-(y[j]+dy[j]))*yconv);
         }
         //printf("xloc1 is %d xloc2 %d yloc1 is %d yloc2 %d\n",xloc1,xloc2,yloc1,yloc2);
         MPE_Draw_line(window, xloc1, yloc2, xloc2, yloc2, MPE_BLACK);
         MPE_Draw_line(window, xloc1, yloc1, xloc2, yloc1, MPE_BLACK);
         MPE_Draw_line(window, xloc1, yloc1, xloc1, yloc2, MPE_BLACK);
         MPE_Draw_line(window, xloc2, yloc1, xloc2, yloc2, MPE_BLACK);

         //MPE_Fill_rectangle(window, xloc, yloc, xwid, ywid, MPE_RED);
      }
   }
#endif

#ifdef HAVE_OPENGL
   /* Draw circle for initial conditions */
   if (display_circle_radius > 0.0) {
      double radians;
      glBegin(GL_LINE_LOOP);
      for (double angle = 0.0; angle <= 360.0; angle+=1.0) {
         radians=angle*3.14159/180.0;
         glVertex2f(sin(radians) * display_circle_radius,cos(radians) * display_circle_radius);
      }
      glEnd();
   }
#endif
  
#ifdef XXX
#ifdef HAVE_OPENGL
   /*Trace order of cells with line going from center to center*/
   glBegin(GL_LINE_STRIP);
      for(j = 0; j < display_jsize; j++) {
         for(i = 0; i < display_isize; i++) {
            if (cell_centered) {
               glVertex2f(x[i], y[j]);
            } else {
               glVertex2f(x[i]+0.5*dx[i], y[j]+0.5*dy[j]);
            }
         }
      }
   glEnd();
#endif
#ifdef HAVE_MPE
   int xloc_old, yloc_old;
   if (cell_centered) {
      xloc_old = (int)((x[0]-display_xmin)*xconv);
      yloc_old = (int)((y[0]-display_ymin)*yconv);
   } else {
      xloc_old = (int)((x[0]+0.5*dx[0]-display_xmin)*xconv);
      yloc_old = (int)((y[0]+0.5*dy[0]-display_ymin)*yconv);
   }
   for(j = 1; j < display_jsize; j++) {
      for(i = 1; i < display_isize; i++) {
         if (cell_centered) {
            xloc = (int)((x[i]-display_xmin)*xconv);
            xloc = (int)((x[i]+0.5*dx[i]-display_xmin)*xconv);
            yloc = (int)((y[j]+0.5*dy[j]-display_ymin)*yconv);
         }
         MPE_Draw_line(window,xloc_old,yloc_old,xloc,yloc,MPE_BLACK);
         xloc_old = xloc;
         yloc_old = yloc;
      }
   }
#endif
#endif

#ifdef HAVE_MPE
#ifdef HAVE_MPI
   MPI_Barrier(MPI_COMM_WORLD);
#endif
   MPE_Update(window);
#endif
}

void DrawBoxes(void) {

#ifdef HAVE_OPENGL
   int j, i, color, step = NCOLORS/(display_proc[display_jsize-1][display_isize-1]+1);
   //step utilizes whole range of color, assumes last element of proc is last processor

   glFrustum(display_xmin-1, display_xmax, display_ymin-1, display_ymax, 5.0, 10.0);
   glTranslatef(0.0f, 0.0f, -6.0f); //must move into screen to draw

   for(j = 0; j < display_jsize; j++) {
      for(i = 0; i < display_isize; i++) {
         /*Draw filled cells with color set by processor*/
         glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
         glBegin(GL_QUADS);

            color = display_proc[j][i]*step;
            glColor3f(Rainbow[color].Red, Rainbow[color].Green, Rainbow[color].Blue);

            /*Front Face*/
            glVertex3f(x[i],       y[j],       0.0f);
            glVertex3f(x[i]+dx[i], y[j],       0.0f);
            glVertex3f(x[i]+dx[i], y[j]+dy[j], 0.0f);
            glVertex3f(x[i],       y[j]+dy[j], 0.0f);
            /*Right Face*/
            glVertex3f(x[i]+dx[i], y[j],       0.0f);
            glVertex3f(x[i]+dx[i], y[j]+dy[j], 0.0f);
            glVertex3f(x[i]+dx[i], y[j]+dy[j],-1.0f);
            glVertex3f(x[i]+dx[i], y[j],      -1.0f);
            /*Back Face*/
            glVertex3f(x[i]+dx[i], y[j],      -1.0f);
            glVertex3f(x[i]+dx[i], y[j]+dy[j],-1.0f);
            glVertex3f(x[i],       y[j]+dy[j],-1.0f);
            glVertex3f(x[i],       y[j],      -1.0f);
            /*Left Face*/
            glVertex3f(x[i],       y[j],       0.0f);
            glVertex3f(x[i],       y[j],      -1.0f);
            glVertex3f(x[i],       y[j]+dy[j],-1.0f);
            glVertex3f(x[i],       y[j]+dy[j], 0.0f);
            /*Bottom*/
            glVertex3f(x[i],       y[j],       0.0f);
            glVertex3f(x[i],       y[j],      -1.0f);
            glVertex3f(x[i]+dx[i], y[j],      -1.0f);
            glVertex3f(x[i]+dx[i], y[j],       0.0f);
            /*Top*/
            glVertex3f(x[i],       y[j]+dy[j], 0.0f);
            glVertex3f(x[i],       y[j]+dy[j],-1.0f);
            glVertex3f(x[i]+dx[i], y[j]+dy[j],-1.0f);
            glVertex3f(x[i]+dx[i], y[j]+dy[j], 0.0f);
         glEnd();
      }
   }
#endif
}



void set_viewmode(int display_view_mode_in){
   display_view_mode = display_view_mode_in;
}
void set_my_jsize(int display_jsize_in){
   display_jsize = display_jsize_in;
}
void set_my_isize(int display_isize_in){
   display_isize = display_isize_in;
}
void set_my_offset(int display_offset_in){
   display_myoffset = display_offset_in;
}
void set_circle_radius(double display_circle_radius_in){
   display_circle_radius = display_circle_radius_in;
}
void set_outline(int display_outline_in){
   display_outline = display_outline_in;
}

void draw_scene(void) {
   draw_contour();
   draw_graph();
   draw_scale();
   draw_xcut();
   draw_status();
#if HAVE_OPENGL
   glutSetWindow(contour_window);
#endif
}

void draw_contour(void) {
#ifdef HAVE_OPENGL
   if (rank) return; // Only processor 0 draws to screen for OpenGL		 
   glutSetWindow(contour_window);

   glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
   glLoadIdentity();
#endif
#ifdef HAVE_MPE
   if (rank == 0) {
      MPE_Fill_rectangle(window, 0, 0, WINSIZE, WINSIZE, MPE_WHITE);
   }
#ifdef HAVE_MPI
   MPI_Barrier(MPI_COMM_WORLD);
#endif
   MPE_Update(window);
#endif

   if (display_view_mode == 0) {
      DrawSquares();
   } else {
      DisplayState();
   }

#ifdef HAVE_OPENGL
#ifdef XXX
   char c[10];

   int icount=0;
   for(int j = 0; j < display_jsize; j++) {
      for(int i = 0; i < display_isize; i++) {
         sprintf(c, "%d", icount);
         if (cell_centered) {
            DrawString(x[i], y[j], 0.0, c);
         } else {
            DrawString(x[i]+0.5*dx[i], y[j]+0.5*dy[j], 0.0, c);
         }
         icount++;
      }
   }
#endif
   if(mode == DRAW) {
      SelectionRec();
   }

   glLoadIdentity();

   glutSwapBuffers();
#endif

   sleep(0);
}


#ifdef HAVE_OPENGL
void SelectionRec(void) {
   glColor3f(0.0f, 0.0f, 0.0f);
   glLineWidth(2.0f);
   glBegin(GL_QUADS);
      glVertex2f(xstart, ystart);
      glVertex2f(xstart, yend);
      glVertex2f(xend, yend);
      glVertex2f(xend, ystart);
   glEnd();
   glLineWidth(1.0f);
}
#endif

void mouseClick(int button, int state, int x, int y) {
   //printf("DEBUG mouse click is button %d state %d xloc %d yloc %d\n",button,state,x,y);
#ifdef HAVE_OPENGL
   if(state == GLUT_DOWN) {
      mode = EYE;
      xstart = (display_xmax-display_ymin)*(x/width)+display_ymin;
      ystart = (display_ymax-display_ymin)*((float)(WINSIZE-y)/WINSIZE)+display_ymin;
   }
   if(state == GLUT_UP) {
      glutPostRedisplay();
      mode = DRAW;
      xend = (display_xmax-display_ymin)*(x/width)+display_ymin;
      yend = (display_xmax-display_ymin)*((float)(WINSIZE-y)/WINSIZE)+display_ymin;
   }
#endif

   mouse_click_function(button, state, x, y);
}
void mouseDrag(int x, int y) {
#ifdef HAVE_OPENGL
   glutPostRedisplay();
   mode = DRAW;
   xend = (display_xmax-display_ymin)*(x/width)+display_ymin;
   yend = (display_xmax-display_ymin)*((float)(WINSIZE-y)/WINSIZE)+display_ymin;
#endif
}

void keyPressed(unsigned char key, int x, int y) {
    if(key == 'm')   { mode = MOVE; }
    if(key == 'e')   { mode = EYE; }
    if(mode == EYE) {
      if(key == 'o') { xrot -= 5.0; }
      if(key == 'l') { xrot += 5.0; }
      if(key == 'k') { yrot -= 5.0; }
      if(key == ';') { yrot += 5.0; }
    }
    if(mode == MOVE) {
      if(key == 'o') { zloc += 1.0; }
      if(key == 'l') { zloc -= 1.0; }
      if(key == 'k') { xloc += 1.0; }
      if(key == ';') { xloc -= 1.0; }
    }

    keyboard_function(key, x, y);
}
void Scale() {
#ifdef HAVE_OPENGL
   int i;
   double r;
   for (i=0, r=0.0; i<200; i++, r+=.005) {
         Rainbow[i].Red = 0.0;
         Rainbow[i].Green = r;
         Rainbow[i].Blue = 1.0;
   }
   for (i=0, r=1.0; i<200; i++, r-=.005) {
         Rainbow[200+i].Red = 0.0;
         Rainbow[200+i].Green = 1.0;
         Rainbow[200+i].Blue = r;
   }
   for (i=0, r=0.0; i<200; i++, r+=.005) {
         Rainbow[400+i].Red = r;
         Rainbow[400+i].Green = 1.0;
         Rainbow[400+i].Blue = 0.0;
   }
   for (i=0, r=1.0; i<200; i++, r-=.005) {
         Rainbow[600+i].Red = 1.0;
         Rainbow[600+i].Green = r;
         Rainbow[600+i].Blue = 0.0;
   }
   for (i=0, r=0.0; i<200; i++, r+=.005) {
         Rainbow[800+i].Red = 1.0;
         Rainbow[800+i].Green = 0.0;
         Rainbow[800+i].Blue = r;
   }
#endif
#ifdef HAVE_MPE
   color_array = (MPE_Color *) malloc(sizeof(MPE_Color)*ncolors);

   int ierr = MPE_Make_color_array(window, ncolors, color_array);
   if (ierr && rank == 0) printf("Error(Make_color_array): ierr is %d\n",ierr);

#endif
}

void mpe_main_loop(void)
{
#ifdef HAVE_MPE
   while (1) {
      display_get_event();
      idlefunction();
   }
#endif
}

/********************************************************************************/
void display_get_event(void)
/********************************************************************************/
{

#ifdef HAVE_MPE
   //double xmid, ymid;
   //int xrel, yrel;
   XEvent event;
   long EventMask;
   int EventFlag;
   char keys[20];
   int numChar;
   KeySym keysym;
   XComposeStatus compose;
   int rank=0;
   unsigned char key;
   int button, xloc, yloc, special_event;
   double xcor, ycor;


#ifdef HAVE_MPI
   MPI_Comm_rank(MPI_COMM_WORLD, &rank);
#endif

   if (window->Cookie != MPE_G_COOKIE) {
      printf("Handle argument is incorrect or corrupted\n" );
   }

   key = '\0';
   button = -1;
   xcor = 0.0;
   ycor = 0.0;
   special_event = 0;

   EventMask= ButtonPressMask | ButtonReleaseMask | KeyPressMask | ExposureMask | StructureNotifyMask ;


   EventFlag=0;
   //if (rank ==0){
      if(XCheckWindowEvent(window->xwin->disp,window->xwin->win,EventMask,&event)){
         EventFlag=1;
      }
   //}
#ifdef HAVE_MPI
   MPI_Bcast(&EventFlag, 1, MPI_INT, 0, MPI_COMM_WORLD);
#endif
     
   if (EventFlag){
#ifdef HAVE_MPI
      MPI_Bcast(&event, (int)sizeof(XEvent), MPI_BYTE, 0, MPI_COMM_WORLD);
#endif

      //printf("Event type is %d\n",event.type);
      switch(event.type){
         case ButtonPress:
            //printf("Button Press is %d\n",event.xbutton.button);
            button=event.xbutton.button;
            xloc=event.xbutton.x;
            yloc=event.xbutton.y;
            xcor=display_xmin+(double)xloc/xconv;
            ycor=display_ymin+(double)(height-yloc -1)/yconv;

            int state = 1; // Button Down
            mouseClick((int)event.xbutton.button, state, xloc, yloc);

            //printf("DEBUG graphics -- button is %d loc is %d %d cor is %lf %lf\n",
            //                       button,    xloc,yloc,xcor,ycor);
            break;
         case KeyPress:
            if (rank == 0) {
               numChar=XLookupString((XKeyEvent *) &event, keys, 20, &keysym, &compose);
               printf("%c:%c:%d:\n",keys[0],keys[1],(int)keysym);
               if(((keysym>=XK_KP_Space)&&(keysym<=XK_KP_9))
                  ||((keysym>XK_space)&&(keysym<XK_asciitilde))){
                  key=keys[0];
               }
            }
#ifdef HAVE_MPI
            MPI_Bcast(&key,           1, MPI_CHAR,   0, MPI_COMM_WORLD);
#endif
     
            xloc=event.xkey.x;
            yloc=event.xkey.y;
     
            xcor=display_xmin+(double)xloc/xconv;
            ycor=display_ymin+(double)(height-yloc-1)/yconv;

            keyboard_function(key, xloc, yloc);
     
            //printf("DEBUG graphics -- key is '%c' loc is %d %d cor is %lf %lf\n",
            //                         key,      xloc,yloc,xcor,ycor);
            break;
         case Expose:
            if (event.xexpose.count == 0){
               special_event=1;
            }
            //printf("DEBUG graphics -- special event is \n",*special_event);
            break;
         case ConfigureNotify:
            /* Window has been resized */
/*
            width=event.xconfigure.width; //-SCALEWIDTH;
            double ratio = (double)event.xconfigure.height/(double)height;
            height *= ratio;
     
            conHeight *= ratio;
            conWidth  = width - SCALEWIDTH;
     
            graHeight = height - conHeight - 2*TEXTHEIGHT;
            if (graHeight < 0) graHeight  = 0;
            graWidth  = width - SCALEWIDTH;
     
            xconv = (double)conWidth/ (display_xmax-xmin);
            yconv = (double)conHeight/(display_ymax-display_ymin);
*/

            //printf("DEBUG graphics -- window has been resized to %d %d\n",width, height);
            break;
      } // switch
     
      //printf("DEBUG -- button is %d key is '%c' loc is %d %d cor is %lf %lf\n",
      //                button,     key,      xloc,yloc,xcor,ycor);
   }
     
   //MPE_Update(window);
   //usleep(300000);
     
#endif

   return;
}
   

