#ifndef _DISPLAY_H_
#define _DISPLAY_H_

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#ifdef HAVE_OPENGL
#ifdef __APPLE_CC__
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/glu.h>
#else
#include <GL/glut.h>
#include <GL/gl.h>
#include <GL/glu.h>
#endif
#endif

#ifdef __cplusplus
extern "C"
{
#endif

#define OMPI_SKIP_MPICXX 1

#ifdef HAVE_MPE
#ifdef HAVE_MPI
#include <mpi.h>
#else
#define MPE_NOMPI
#define CLOG_NOMPI
#endif
#define MPE_INTERNAL
#define MPE_GRAPHICS
#include "mpe.h"
#define MPE_DRAG_FIXED_RECT 8
#include <X11/Xlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
//keyboard input files
#include <termios.h>
#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <math.h>
#endif

void init_display(int *argc, char **argv, const char* string, int mype_in);
void set_display_label(char *display_label_in);
void set_idle_function(void (*function)(void));
void set_keyboard_function(void (*function)(unsigned char, int, int));
void set_mouse_click_function(void (*function)(int, int, int, int));
void start_main_loop(void);
void free_display(void);
void set_viewmode(int display_view_mode_in);
void set_outline(int display_outline_in);
void set_my_isize(int my_isize_in);
void set_my_jsize(int my_jsize_in);
void set_my_offset(int my_offset_in);
void set_window(double display_xmin_in, double display_xmax_in, double display_ymin_in, double display_ymax_in);
void set_cell_coordinates(double *x_in, double *dx_in, double *y_in, double *dy_in);
void set_cell_data(double **data_in);
void set_cell_proc(int **display_proc_in);
void set_circle_radius(double display_circle_radius_in);
void draw_scene(void);
void draw_contour(void);
void draw_graph(void);
void draw_scale(void);
void draw_xcut(void);
void draw_status(void);

#ifdef __cplusplus
}
#endif

#endif

