#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "display.h"
#include "genmalloc/genmalloc.h"

typedef unsigned int uint;

extern "C" void do_calc(void);
extern "C" void mykeyPressed(unsigned char key, int x, int y);
extern "C" void mymouseClick(int button, int state, int x, int y);

static int view_mode = 1;

int         outputInterval=100, //  Periodicity of output
            outline = 1,
            niter=10000;        //  Maximum time step

static double **display_var;
static double *x, *y, *dx, *dy;
uint imax, jmax;

int main(int argc, char **argv) {
   int mype=0;

   double deltax = 1.0, deltay = 1.0;
   
   imax = 128; jmax = 128;

   double xmin = 0.0, xmax = 128.0;
   double ymin = 0.0, ymax = 128.0;

   dx = (double *)malloc(imax*sizeof(double));
   for (uint i=0; i<imax; i++){
      dx[i]=deltax;
   }
   dy = (double *)malloc(jmax*sizeof(double));
   for (uint j=0; j<jmax; j++){
      dy[j]=deltay;
   }
   x = (double *)malloc(imax*sizeof(double));
   x[0] = 0.0;
   for (uint i=1; i<imax; i++){
      x[i]=x[i-1]+dx[i];
   }
   y = (double *)malloc(jmax*sizeof(double));
   y[0] = 0.0;
   for (uint j=1; j<jmax; j++){
      y[j]=y[j-1]+dy[j];
   }

   int **proc = (int **)genmatrix(jmax, imax, sizeof(int));
   for (uint j=0; j<jmax; j++){
      for (uint i=0; i<imax; i++){
         proc[j][i] = 0;
      }
   }

   printf("Iteration:%8d, Time:%8.2f, Timestep: n/a\n", 0, 0.0);

   display_var = (double **)genmatrix(jmax, imax, sizeof(double));
   for (uint j=0; j<jmax; j++){
      for (uint i=0; i<imax; i++){
         display_var[j][i] = ((double)j + (double)i)/50.0;
      }
   }

#ifdef HAVE_GRAPHICS
   set_window(xmin, xmax, ymin, ymax);
   set_my_isize(imax);
   set_my_jsize(jmax);
   set_cell_proc(proc);
   set_viewmode(view_mode);
   set_outline(outline);
   set_cell_coordinates(&x[0], &dx[0], &y[0], &dy[0]);
   set_circle_radius(NULL);
   set_cell_data(display_var);
   void (*ptrFunc)(unsigned char, int, int);
   ptrFunc=mykeyPressed;
   set_keyboard_function(ptrFunc);
   void (*ptrFunc2)(int, int, int, int);
   ptrFunc2=mymouseClick;
   set_mouse_click_function(ptrFunc2);
   init_display(&argc, argv, "Visualization Test", mype);
   draw_scene();

   set_idle_function(&do_calc);
   start_main_loop();
#else
   for (int it = 0; it < 10000000; it++) {
      do_calc();
   }
#endif
   
   return 0;
}

static int     ncycle  = 0;
static double  simTime = 0.0;
static double  deltaT  = 0.001;

extern "C" void do_calc(void)
{  
   for (int nburst = 0; nburst < outputInterval && ncycle < niter; nburst++, ncycle++) {
      for (uint j=0; j<jmax; j++){
         for (uint i=0; i<imax; i++){
            display_var[j][i] += 0.1;
            if (display_var[j][i] > 100.0) display_var[j][i] = ((double)j + (double)i)/10.0;
         }
      }
      simTime += deltaT;
   } // End of iteration loop

   printf("Iteration:%8d, Time:%8.4f, Timestep: %.4f\n", ncycle, simTime, deltaT);

#ifdef HAVE_GRAPHICS
   set_viewmode(view_mode);
   set_cell_proc(NULL);
   draw_scene();
   //sleep(1);
   usleep(10);
#endif

   if (ncycle >= niter) exit(0);
}

#define ESCAPE 27
void mykeyPressed(unsigned char key, int x, int y) {
    printf("DEBUG key pressed is %d\n",key);

    sleep(1);

    if(key == ESCAPE) {
       exit(0);
    }
}

void mymouseClick(int button, int state, int x, int y) {
    printf("DEBUG mouse click is button %d state %d xloc %d yloc %d\n",button, state, x, y);

    sleep(1);
}
